import logging
import os
import sys

from .common import RankTableEnv, setup_logger
from .fmk import FMKManager
from .rank_table import RankTable, RankTableTemplate1, RankTableTemplate2

logger = logging.getLogger(__name__)


def app():
    if len(sys.argv) <= 1:
        logger.error("Usage: ma2l YOUR_TRAINING_COMMAND")
        sys.exit(1)

    setup_logger(__package__)
    train_command = sys.argv[1:]
    logger.info(f"training command: {' '.join(train_command)}")

    if os.environ.get(RankTableEnv.RANK_TABLE_FILE_V1) is not None:
        # new format rank table file
        rank_table_path = os.environ.get(RankTableEnv.RANK_TABLE_FILE_V1)
        RankTable.wait_for_available(rank_table_path)
        rank_table = RankTableTemplate1(rank_table_path)
    else:
        # old format rank table file
        rank_table_path_origin = RankTableEnv.get_rank_table_template2_file_path()
        RankTable.wait_for_available(rank_table_path_origin)
        rank_table = RankTableTemplate2(rank_table_path_origin)

    if rank_table.get_device_num() >= 1:
        logger.info(f"set rank table {RankTableEnv.RANK_TABLE_FILE} env to {rank_table.get_rank_table_path()}")
        RankTableEnv.set_rank_table_env(rank_table.get_rank_table_path())
    else:
        logger.info(f"device num < 1, unset rank table {RankTableEnv.RANK_TABLE_FILE} env")
        RankTableEnv.unset_rank_table_env()

    instance = rank_table.get_current_instance()
    server = rank_table.get_server(instance.server_id)
    current_instance = RankTable.convert_server_to_instance(server)

    fmk_manager = FMKManager(current_instance)
    fmk_manager.run(rank_table.get_device_num(), train_command)
    return_code = fmk_manager.monitor()

    fmk_manager.destroy()

    sys.exit(return_code)


def main():
    app()



import logging
import os
import sys


class RankTableEnv:
    RANK_TABLE_FILE = "RANK_TABLE_FILE"
    RANK_TABLE_FILE_V1 = "RANK_TABLE_FILE_V_1_0"

    HCCL_CONNECT_TIMEOUT = "HCCL_CONNECT_TIMEOUT"

    # jobstart_hccl.json is provided by the volcano controller of Cloud-Container-Engine(CCE)
    HCCL_JSON_FILE_NAME = "jobstart_hccl.json"
    RANK_TABLE_FILE_DEFAULT_VALUE = f"/user/config/{HCCL_JSON_FILE_NAME}"

    @staticmethod
    def get_rank_table_template1_file_dir():
        parent_dir = os.environ[ModelArts.MA_MOUNT_PATH_ENV]
        return os.path.join(parent_dir, "rank_table")

    @staticmethod
    def get_rank_table_template2_file_path():
        rank_table_file_path = os.environ.get(RankTableEnv.RANK_TABLE_FILE)
        if rank_table_file_path is None:
            return RankTableEnv.RANK_TABLE_FILE_DEFAULT_VALUE
        return os.path.join(os.path.normpath(rank_table_file_path), RankTableEnv.HCCL_JSON_FILE_NAME)

    @staticmethod
    def set_rank_table_env(path):
        os.environ[RankTableEnv.RANK_TABLE_FILE] = path

    @staticmethod
    def unset_rank_table_env():
        del os.environ[RankTableEnv.RANK_TABLE_FILE]


class ModelArts:
    MA_MOUNT_PATH_ENV = "MA_MOUNT_PATH"
    MA_CURRENT_INSTANCE_NAME_ENV = "MA_CURRENT_INSTANCE_NAME"
    MA_VJ_NAME = "MA_VJ_NAME"

    MA_CURRENT_HOST_IP = "MA_CURRENT_HOST_IP"

    CACHE_DIR = "/cache"

    TMP_LOG_DIR = "/tmp/log/"

    FMK_WORKSPACE = "workspace"

    @staticmethod
    def get_current_instance_name():
        return os.environ[ModelArts.MA_CURRENT_INSTANCE_NAME_ENV]

    @staticmethod
    def get_current_host_ip():
        return os.environ.get(ModelArts.MA_CURRENT_HOST_IP)

    @staticmethod
    def get_job_id():
        ma_vj_name = os.environ[ModelArts.MA_VJ_NAME]
        return ma_vj_name.replace("ma-job", "modelarts-job", 1)

    @staticmethod
    def get_parent_working_dir():
        if ModelArts.MA_MOUNT_PATH_ENV in os.environ:
            return os.path.join(os.environ.get(ModelArts.MA_MOUNT_PATH_ENV), ModelArts.FMK_WORKSPACE)
        return ModelArts.CACHE_DIR


def setup_logger(name):
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    logger.propagate = False
    fmt = "%(asctime)s %(name)s %(levelname)s - %(message)s"
    datefmt = "[%Y-%m-%d %H:%M:%S]"
    console_handler = logging.StreamHandler(stream=sys.stdout)
    console_handler.setFormatter(logging.Formatter(fmt=fmt, datefmt=datefmt))
    logger.addHandler(console_handler)
    return logger



import json
import logging
import os
import time

from .common import ModelArts, RankTableEnv

logger = logging.getLogger(__name__)


class Device:
    def __init__(self, device_id, device_ip, rank_id):
        self.device_id = device_id
        self.device_ip = device_ip
        self.rank_id = rank_id


class Instance:
    def __init__(self, pod_name, server_id, devices):
        self.pod_name = pod_name
        self.server_id = server_id
        self.devices = self.parse_devices(devices)

    @staticmethod
    def parse_devices(devices):
        if devices is None:
            return []

        device_object_list = []
        for device in devices:
            device_object_list.append(Device(device["device_id"], device["device_ip"], ""))
        return device_object_list

    def set_devices(self, devices):
        self.devices = devices


class Group:
    def __init__(self, group_name, device_count, instance_count, instance_list):
        self.group_name = group_name
        self.device_count = int(device_count)
        self.instance_count = int(instance_count)
        self.instance_list = self.parse_instance_list(instance_list)

    @staticmethod
    def parse_instance_list(instance_list):
        instance_object_list = []
        for instance in instance_list:
            instance_object_list.append(Instance(instance["pod_name"], instance["server_id"], instance["devices"]))

        return instance_object_list


class RankTable:
    STATUS_FIELD = "status"
    COMPLETED_STATUS = "completed"

    def __init__(self):
        self.rank_table_path = ""
        self.rank_table = {}

    @staticmethod
    def read_from_file(file_path):
        with open(file_path) as json_file:
            return json.load(json_file)

    @staticmethod
    def wait_for_available(rank_table_file, period=1):
        logger.info(f"Wait for Rank table file at {rank_table_file} ready")
        complete_flag = False
        while not complete_flag:
            with open(rank_table_file) as json_file:
                data = json.load(json_file)
                if data[RankTable.STATUS_FIELD] == RankTable.COMPLETED_STATUS:
                    logger.info("Rank table file is ready for read")
                    logger.info("\n" + json.dumps(data, indent=4))
                    return True

            time.sleep(period)

        return False

    @staticmethod
    def convert_server_to_instance(server):
        device_list = []
        for device in server["device"]:
            device_list.append(
                Device(device_id=device["device_id"], device_ip=device["device_ip"], rank_id=device["rank_id"])
            )

        ins = Instance(pod_name="", server_id=server["server_id"], devices=[])
        ins.set_devices(device_list)
        return ins

    def get_rank_table_path(self):
        return self.rank_table_path

    def get_server(self, server_id):
        for server in self.rank_table["server_list"]:
            if server["server_id"] == server_id:
                logger.info("Current server")
                logger.info("\n" + json.dumps(server, indent=4))
                return server

        logger.error(f"server [{server_id}] is not found")
        return None


class RankTableTemplate1(RankTable):
    def __init__(self, rank_table_template1_path):
        super().__init__()
        self.rank_table_path = rank_table_template1_path
        self.rank_table = self.read_from_file(file_path=rank_table_template1_path)

    def get_current_instance(self):
        current_server = None
        server_list = self.rank_table["server_list"]
        if len(server_list) == 1:
            current_server = server_list[0]
        elif len(server_list) > 1:
            host_ip = ModelArts.get_current_host_ip()
            if host_ip is not None:
                for server in server_list:
                    if server["server_id"] == host_ip:
                        current_server = server
                        break
            else:
                current_server = server_list[0]

        if current_server is None:
            logger.error("server is not found")
            return None
        return self.convert_server_to_instance(current_server)

    def get_device_num(self):
        server_list = self.rank_table["server_list"]
        device_num = 0
        for server in server_list:
            device_num += len(server["device"])
        return device_num


class RankTableTemplate2(RankTable):
    def __init__(self, rank_table_template2_path):
        super().__init__()

        json_data = self.read_from_file(file_path=rank_table_template2_path)

        self.status = json_data[RankTableTemplate2.STATUS_FIELD]
        if self.status != RankTableTemplate2.COMPLETED_STATUS:
            return

        # sorted instance list by the index of instance
        # assert there is only one group
        json_data["group_list"][0]["instance_list"] = sorted(
            json_data["group_list"][0]["instance_list"],
            key=RankTableTemplate2.get_index,
        )

        self.group_count = int(json_data["group_count"])
        self.group_list = self.parse_group_list(json_data["group_list"])

        self.rank_table_path, self.rank_table = self.convert_template2_to_template1_format_file()

    @staticmethod
    def parse_group_list(group_list):
        group_object_list = []
        for group in group_list:
            group_object_list.append(
                Group(group["group_name"], group["device_count"], group["instance_count"], group["instance_list"])
            )

        return group_object_list

    @staticmethod
    def get_index(instance):
        # pod_name example: job94dc1dbf-job-bj4-yolov4-15
        pod_name = instance["pod_name"]
        return int(pod_name[pod_name.rfind("-") + 1 :])

    def get_current_instance(self):
        """
        get instance by pod name
        specially, return the first instance when the pod name is None
        :return:
        """
        pod_name = ModelArts.get_current_instance_name()
        if pod_name is None:
            if len(self.group_list) > 0:
                if len(self.group_list[0].instance_list) > 0:
                    return self.group_list[0].instance_list[0]

            return None

        for group in self.group_list:
            for instance in group.instance_list:
                if instance.pod_name == pod_name:
                    return instance
        return None

    def convert_template2_to_template1_format_file(self):
        rank_table_template1_file = {
            "status": "completed",
            "version": "1.0",
            "server_count": "0",
            "server_list": [],
        }

        logic_index = 0
        server_map = {}
        # collect all devices in all groups
        for group in self.group_list:
            if group.device_count == 0:
                continue
            for instance in group.instance_list:
                if instance.server_id not in server_map:
                    server_map[instance.server_id] = []

                for device in instance.devices:
                    template1_device = {
                        "device_id": device.device_id,
                        "device_ip": device.device_ip,
                        "rank_id": str(logic_index),
                    }
                    logic_index += 1
                    server_map[instance.server_id].append(template1_device)

        server_count = 0
        for server_id in server_map:
            rank_table_template1_file["server_list"].append({"server_id": server_id, "device": server_map[server_id]})
            server_count += 1

        rank_table_template1_file["server_count"] = str(server_count)

        logger.info("Rank table file (Template1)")
        logger.info("\n" + json.dumps(rank_table_template1_file, indent=4))

        if not os.path.exists(RankTableEnv.get_rank_table_template1_file_dir()):
            os.makedirs(RankTableEnv.get_rank_table_template1_file_dir())

        path = os.path.join(RankTableEnv.get_rank_table_template1_file_dir(), RankTableEnv.HCCL_JSON_FILE_NAME)
        with open(path, "w") as f:
            f.write(json.dumps(rank_table_template1_file))
            logger.info(f"Rank table file (Template1) is generated at {path}")

        return path, rank_table_template1_file

    def get_device_num(self):
        total_device_num = 0
        for group in self.group_list:
            total_device_num += group.device_count
        return total_device_num



import logging
import os
import pathlib
import signal
import subprocess
import time
from contextlib import contextmanager

from .common import ModelArts, RankTableEnv

logger = logging.getLogger(__name__)


class FMK:
    def __init__(self, index, device):
        self.job_id = ModelArts.get_job_id()
        self.rank_id = device.rank_id
        self.device_id = str(index)

    def gen_env_for_fmk(self, rank_size):
        current_envs = os.environ.copy()

        current_envs["JOB_ID"] = self.job_id

        current_envs["ASCEND_DEVICE_ID"] = self.device_id
        current_envs["DEVICE_ID"] = self.device_id

        current_envs["RANK_ID"] = self.rank_id
        current_envs["RANK_SIZE"] = str(rank_size)

        FMK.set_env_if_not_exist(current_envs, RankTableEnv.HCCL_CONNECT_TIMEOUT, str(1800))

        log_dir = FMK.get_log_dir()
        process_log_path = os.path.join(log_dir, self.job_id, "ascend", "process_log", f"rank_{self.rank_id}")
        FMK.set_env_if_not_exist(current_envs, "ASCEND_PROCESS_LOG_PATH", process_log_path)
        pathlib.Path(current_envs["ASCEND_PROCESS_LOG_PATH"]).mkdir(parents=True, exist_ok=True)

        return current_envs

    @contextmanager
    def switch_directory(self, directory):
        owd = os.getcwd()
        try:
            os.chdir(directory)
            yield directory
        finally:
            os.chdir(owd)

    def get_working_dir(self):
        fmk_workspace_prefix = ModelArts.get_parent_working_dir()
        return os.path.join(os.path.normpath(fmk_workspace_prefix), f"device{self.device_id}")

    @staticmethod
    def get_log_dir():
        parent_path = os.getenv(ModelArts.MA_MOUNT_PATH_ENV)
        if parent_path:
            log_path = os.path.join(parent_path, "log")
            if os.path.exists(log_path):
                return log_path

        return ModelArts.TMP_LOG_DIR

    @staticmethod
    def set_env_if_not_exist(envs, env_name, env_value):
        if env_name in os.environ:
            logger.info(f"env already exists. env_name: {env_name}, env_value: {env_value}")
            return

        envs[env_name] = env_value

    def run(self, rank_size, command):
        envs = self.gen_env_for_fmk(rank_size)
        proc = f"proc-rank-{self.rank_id}-device-{self.device_id}"
        logger.info(f"bootstrap {proc}")

        log_dir = FMK.get_log_dir()
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)

        log_file = f"{self.job_id}-{proc}.txt"
        log_file_path = os.path.join(log_dir, log_file)

        # os.setsid: change the process(forked) group id to itself
        training_proc = subprocess.Popen(
            command,
            env=envs,
            preexec_fn=os.setsid,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        logger.info(f"{proc} (pid: {training_proc.pid})")
        # https://docs.python.org/3/library/subprocess.html#subprocess.Popen.wait
        subprocess.Popen(["tee", log_file_path], stdin=training_proc.stdout)

        return training_proc


class FMKManager:
    # max destroy time: ~20 (15 + 5)
    # ~ 15 (1 + 2 + 4 + 8)
    MAX_TEST_PROC_CNT = 4

    def __init__(self, instance):
        self.instance = instance
        self.fmk = []
        self.fmk_processes = []
        self.get_sigterm = False
        self.max_test_proc_cnt = FMKManager.MAX_TEST_PROC_CNT

    # break the monitor and destroy processes when get terminate signal
    def term_handle(func):
        def receive_term(signum, stack):
            logger.info(f"Received terminate signal {signum}, try to destroyed all processes")
            stack.f_locals["self"].get_sigterm = True

        def handle_func(self, *args, **kwargs):
            origin_handle = signal.getsignal(signal.SIGTERM)
            signal.signal(signal.SIGTERM, receive_term)
            res = func(self, *args, **kwargs)
            signal.signal(signal.SIGTERM, origin_handle)
            return res

        return handle_func

    def run(self, rank_size, command):
        for index, device in enumerate(self.instance.devices):
            fmk_instance = FMK(index, device)
            self.fmk.append(fmk_instance)
            self.fmk_processes.append(fmk_instance.run(rank_size, command))

    @term_handle
    def monitor(self, period=1):
        # busy waiting for all fmk processes exit by zero
        # or there is one process exit by non-zero

        fmk_cnt = len(self.fmk_processes)
        zero_ret_cnt = 0
        while zero_ret_cnt != fmk_cnt:
            zero_ret_cnt = 0
            for index in range(fmk_cnt):
                fmk = self.fmk[index]
                fmk_process = self.fmk_processes[index]
                if fmk_process.poll() is not None:
                    if fmk_process.returncode != 0:
                        proc = f"proc-rank-{fmk.rank_id}-device-{fmk.device_id} (pid: {fmk_process.pid})"
                        logger.error(f"{proc} has exited with non-zero code: {fmk_process.returncode}")
                        return fmk_process.returncode

                    zero_ret_cnt += 1
            if self.get_sigterm:
                break
            time.sleep(period)

        return 0

    def destroy(self, base_period=1):
        logger.info("Begin destroy training processes")
        self.send_sigterm_to_fmk_process()
        self.wait_fmk_process_end(base_period)
        logger.info("End destroy training processes")

    def send_sigterm_to_fmk_process(self):
        # send SIGTERM to fmk processes (and process group)
        for r_index in range(len(self.fmk_processes) - 1, -1, -1):
            fmk = self.fmk[r_index]
            fmk_process = self.fmk_processes[r_index]
            if fmk_process.poll() is not None:
                proc = f"proc-rank-{fmk.rank_id}-device-{fmk.device_id} (pid: {fmk_process.pid})"
                logger.info(f"{proc} has exited before receiving the term signal")
                del self.fmk_processes[r_index]
                del self.fmk[r_index]

            try:
                os.killpg(fmk_process.pid, signal.SIGTERM)
            except ProcessLookupError:
                pass

    def wait_fmk_process_end(self, base_period):
        test_cnt = 0
        period = base_period
        while len(self.fmk_processes) > 0 and test_cnt < self.max_test_proc_cnt:
            for r_index in range(len(self.fmk_processes) - 1, -1, -1):
                fmk = self.fmk[r_index]
                fmk_process = self.fmk_processes[r_index]
                if fmk_process.poll() is not None:
                    proc = f"proc-rank-{fmk.rank_id}-device-{fmk.device_id} (pid: {fmk_process.pid})"
                    logger.info(f"{proc} has exited")
                    del self.fmk_processes[r_index]
                    del self.fmk[r_index]
            if not self.fmk_processes:
                break

            time.sleep(period)
            period *= 2
            test_cnt += 1

        if len(self.fmk_processes) > 0:
            for r_index in range(len(self.fmk_processes) - 1, -1, -1):
                fmk = self.fmk[r_index]
                fmk_process = self.fmk_processes[r_index]
                if fmk_process.poll() is None:
                    proc = f"proc-rank-{fmk.rank_id}-device-{fmk.device_id} (pid: {fmk_process.pid})"
                    logger.warning(f"{proc} has not exited within the max waiting time, send kill signal")
                    os.killpg(fmk_process.pid, signal.SIGKILL)
